# Copyright (c) 2020 - 2020 TomTom N.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Mapping
import logging
import sys

import hopic.build

log = logging.getLogger(__name__)


def commisery(volume_vars: Mapping, *, require_ticket: bool = False):
    check_template = {
        'image': None,
        'foreach': 'AUTOSQUASHED_COMMIT',
        'sh': (sys.executable, '-m', 'commisery.checking', '${AUTOSQUASHED_COMMIT}'),
    }

    if require_ticket:
        hopic_git_info = hopic.build.HopicGitInfo.from_repo(volume_vars['WORKSPACE'])

        if hopic_git_info.target_commit:
            check_template = {
                'image': None,
                'sh': (sys.executable, '-m', 'commisery.checking', '-j', f'{hopic_git_info.target_commit}..HEAD'),
            }
        else:
            log.info('Not checking ticket presence in commit messages, since no target was prepared.')

    return [
        check_template,
        {
            'sh': (sys.executable, '-m', 'commisery.checking', 'HEAD'),
        },
    ]



#!/usr/bin/env python3

# Copyright (c) 2019 - 2020 TomTom N.V. (https://tomtom.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This is a simplistic implementation of checking adherance to Conventional Commits https://www.conventionalcommits.org/

from pathlib import Path
from .commit import CommitMessage

from collections import namedtuple
from functools import wraps
from inspect import getfullargspec
from stemming.porter2 import stem
import difflib
import itertools
import re
import regex
import sys
import subprocess
from typing import (
        get_type_hints,
        Iterable,
        Sequence,
    )


def type_check(f):
    @wraps(f)
    def validate_parameters(*args, **kwargs):
        func_args = getfullargspec(f)[0]
        kw = kwargs.copy()
        kw.update(dict(zip(func_args, args)))

        for attr_name, attr_type in get_type_hints(f).items():
            if attr_name == 'return':
                continue

            if attr_name in kw and not isinstance(kw[attr_name], attr_type):
                raise TypeError(f"Argument {attr_name!r} is not of type {attr_type}")
        return f(*args, **kwargs)
    return validate_parameters


MatchGroup = namedtuple('MatchGroup', ('text', 'name', 'start', 'end'))
DEFAULT_ACCEPTED_TAGS = frozenset((
    'build',
    'chore',
    'ci',
    'docs',
    'perf',
    'refactor',
    'revert',
    'style',
    'test',
    'improvement',
))


def check_commit(commit, custom_accepted_tags=None, require_ticket=False):
    try:
        if re.match(r'^[0-9a-fA-F]{40}$', str(commit)):
            raise IOError('a full commit hash should never be treated as a file')

        with open(commit, encoding='UTF-8') as f:
            message = CommitMessage(f.read())
    except IOError:
        commit = subprocess.check_output(('git', 'rev-parse', commit))[:-1].decode('UTF-8')
        message = subprocess.check_output(('git', 'show', '-q', '--format=%B', commit, '--'))[:-1].decode('UTF-8')

    have_error = False
    for error in check_commit_message(commit, message, custom_accepted_tags):
        have_error = True
        print(error, file=sys.stderr)

    if require_ticket and count_ticket_refs(message) == 0:
        have_error = True
        print(f"\x1B[1m{commit}: \x1B[31merror\x1B[39m: no ticket was referenced in the provided commit\x1B[m",
              file=sys.stderr)

    return 1 if have_error else 0


JIRA_RE = re.compile(r'\b(?!' +
                     '|'.join(re.escape(i + '-') for i in (
                         'AES',  # AES-128
                         'PEP',  # PEP-440
                         'SHA',  # SHA-256
                         'UTF',  # UTF-8
                         'VT',  # VT-220
                     )) +
                     r')[A-Z]+-[0-9]+\b')


def count_ticket_refs(message):
    """
    Returns the number of Jira tickets referenced anywhere in the commit message.
    """
    return len(JIRA_RE.findall(message))


def check_commit_message(commit, message, custom_accepted_tags=None):
    if isinstance(message, str):
        message = CommitMessage(message)

    if not message.lines[-1]:
        yield f"\x1B[1m{commit}:{len(message.lines)}:1: \x1B[31merror\x1B[39m: commit message body is followed by empty lines\x1B[m"

    if re.match(r"^Merge (?:branch|tag) '.*?'(?:into '.*')?$", message.subject):
        # Ignore branch/tag merges
        return

    accepted_tags = frozenset(custom_accepted_tags) if custom_accepted_tags is not None else DEFAULT_ACCEPTED_TAGS

    subject_re = re.compile(r'''
        ^
        # 1. Commits MUST be prefixed with a type, which consists of a noun, feat, fix, etc., ...
        (?P<type_tag>\w+)

        # 4. A scope MAY be provided after a type. A scope MUST consist of a noun describing a section of the codebase
        #    surrounded by parenthesis, e.g., `fix(parser):`
        (?: \( (?P<scope> [^()]* ) \) )?

        # 1. Commits MUST be prefixed with a type, ..., followed by ..., OPTIONAL `!`, ...
        (?P<breaking>\s*!(?:\s+(?=:))?)?

        # 1. Commits MUST be prefixed with a type, ..., and REQUIRED terminal colon and space.
        (?P<separator>:?[ ]?)

        # 5. A description MUST immediately follow the colon and space after the type/scope prefix. The description is a
        #    short description of the code changes, e.g., fix: array parsing issue when multiple spaces were contained in
        #    string.
        (?P<description>.*)

        $
        ''', re.VERBOSE)
    subject = subject_re.match(message.subject)
    if not subject:
        error = f"\x1B[1m{commit}:1:1: \x1B[31merror\x1B[39m: commit message's subject not formatted according to Conventional Commits\x1B[m\n{subject_re.pattern}\n"
        error += message.full_subject + '\n'
        error += ' ' * message.subject_start + '\x1B[32m' + '^' * max(len(message.full_subject) - message.subject_start, 1) + '\x1B[39m'
        yield error

    def extract_match_group(match, group, start=0):
        if match is None or match.group(group) is None:
            return None
        return MatchGroup(name=group, text=match.group(group), start=match.start(group)+start, end=match.end(group)+start)

    type_tag    = extract_match_group(subject, 'type_tag'   , message.subject_start)
    scope       = extract_match_group(subject, 'scope'      , message.subject_start)
    breaking    = extract_match_group(subject, 'breaking'   , message.subject_start)
    separator   = extract_match_group(subject, 'separator'  , message.subject_start)
    description = extract_match_group(subject, 'description', message.subject_start)

    @type_check
    def complain_about_excess_space(match: MatchGroup, line: int = 0) -> None:
        excess_whitespace = list(itertools.chain.from_iterable(
                range(*space.span()) for space in re.finditer(r'\s{2,}|^\s+|\s+$', match.text)))
        if excess_whitespace:
            error = "\x1B[1m{commit}:{line}:{col}: \x1B[31merror\x1B[39m: excess whitespace in {match.name}\x1B[m\n".format(line=line + 1, col=match.start + 1 + excess_whitespace[0], match=match, commit=commit)
            error += message.full_subject + '\n'
            error += '\x1B[32m'
            cur = -match.start
            for pos in excess_whitespace:
                error += ' ' * (pos - cur) + '^'
                cur = pos + 1
            error += '\x1B[39m'
            yield error

    # 1. Commits MUST be prefixed with a type, which consists of a noun, feat, fix, etc., followed by a colon and a space.
    # 2. The type feat MUST be used when a commit adds a new feature to your application or library.
    # 3. The type fix MUST be used when a commit represents a bug fix for your application.
    if type_tag and type_tag.text not in accepted_tags and type_tag.text not in ('feat', 'fix'):
        error = f"\x1B[1m{commit}:1:1: \x1B[31merror\x1B[39m: use of type tag that's neither 'feat', 'fix' nor whitelisted ({', '.join(accepted_tags)})\x1B[m\n"
        error += message.full_subject + '\n'
        error += ' ' * type_tag.start + '\x1B[31m' + '~' * (type_tag.end - type_tag.start) + '\x1B[39m'
        possibilities = difflib.get_close_matches(type_tag.text, {'feat', 'fix'} | accepted_tags, n=1)
        if possibilities:
            error += '\n' + possibilities[0]
        yield error

    # 4. An optional scope MAY be provided after a type. A scope is a phrase describing a section of the codebase enclosed
    #    in parenthesis, e.g., fix(parser):
    if scope is not None:
        yield from complain_about_excess_space(scope)

    # 13. If included in the type/scope prefix, breaking changes MUST be indicated by a `!` immediately before the `:`.
    #     If `!` is used, `BREAKING CHANGE: ` MAY be ommitted from the footer section, and the commit description SHALL be
    #     used to describe the breaking change.
    if breaking is not None and breaking.text != '!':
        error = f"\x1B[1m{commit}:1:{breaking.start + 1}: \x1B[31merror\x1B[39m: breaking change indicator in commit message's subject should be exactly '!'\x1B[m\n"
        error += message.full_subject + '\n'
        error += breaking.start * ' ' + '\x1B[31m' + breaking.text.find('!') * '~' + ' ' + (len(breaking.text) - 1 - breaking.text.find('!')) * '~' + '\x1B[39m'
        yield error

    # 1. Commits MUST be prefixed with a type, ..., followed by a colon and a space.
    if separator and separator.text != ': ':
        error = f"\x1B[1m{commit}:1:{separator.start + 1}: \x1B[31merror\x1B[39m: commit message's subject lacks a ': ' separator after the type tag\x1B[m\n"
        error += message.full_subject + '\n'
        error += separator.start * ' ' + '\x1B[32m^' * max(1, separator.end - separator.start) + '\x1B[39m'
        yield error

    # 5. A description MUST immediately follow the type/scope prefix. The description is a short description of the
    #    code changes, e.g., fix: array parsing issue when multiple spaces were contained in string.
    if description and not description.text:
        error = f"\x1B[1m{commit}:1:{description.start + 1}: \x1B[31merror\x1B[39m: commit message's subject lacks a description after the type tag\x1B[m\n"
        error += message.full_subject + '\n'
        error += ' ' * description.start + '\x1B[32m^\x1B[39m'
        yield error

    def complain_about_review_refs(component, component_start=0, lineno=0, quote_text=None):
        """
        Disallow referring to review comments because it's a poor excuse for a proper commit message
        """

        review_comment_ref = None
        words = tuple((word.group(), stem(word.group()).lower(), word.start(), word.end()) for word in re.finditer(r'\b(?:\w|[-])+\b', component))
        opt_prefix = frozenset({
                'all',
                'as',
                'code',
                'edit',
                'for',
                'minor',
                'per',
                'request',
            })
        opt_suffix = frozenset({
                'comment',
                'find',
                'with',
            })
        reference_words = frozenset({
                'accord',
                'bitbucket',
                'address',
                'appli',
                'chang',
                'fix',
                'implement',
                'incorpor',
                'per',
                'process',
                'resolv',
                'rework',
            })
        ref_start, ref_end = None, None
        def encounter(word: str, opts: Sequence[str]) -> bool:
            return bool(word in opts or difflib.get_close_matches(word, opts, cutoff=0.9))
        for idx, (word, stemmed, start, end) in enumerate(words):
            if stemmed not in ('review', 'onlin'):
                continue
            min_idx = idx
            while min_idx > 0 and encounter(words[min_idx-1][1], opt_prefix):
                min_idx -= 1
            if min_idx > 0 and encounter(words[min_idx-1][1], reference_words):
                min_idx -= 1
                ref_start = words[min_idx][2]
                ref_end = end
            max_idx = idx
            while max_idx + 1 < len(words) and encounter(words[max_idx+1][1], opt_suffix):
                max_idx += 1
                if ref_end is not None:
                    ref_end = max(ref_end, words[max_idx][3])
            if max_idx + 1 < len(words) and encounter(words[max_idx+1][1], reference_words):
                max_idx += 1
                if ref_start is None:
                    ref_start = len(component)
                if ref_end is None:
                    ref_end = 0
                ref_start = min(ref_start, words[min_idx][2])
                ref_end = max(ref_end, words[max_idx][3])
            if ref_start is None and ref_end is None:
                brace_prefix = re.match(r'^.*([(])\s*', component[:start])
                brace_suffix = re.match(r'\s*([)])', component[words[max_idx][3]:])
                if brace_prefix and brace_suffix:
                    ref_start = brace_prefix.start(1)
                    ref_end = brace_prefix.end(1)
            if ref_start is None and ref_end is None:
                for try_idx, (try_word, try_stemmed, _, _) in enumerate(words[min_idx:max_idx+1], min_idx):
                    if encounter(try_stemmed, reference_words):
                        ref_start = words[min_idx][2]
                        ref_end = words[max_idx][3]
                        break
            if ref_start is not None and ref_end is not None:
                review_comment_ref = MatchGroup(name=0, text=component[ref_start:ref_end], start=ref_start+component_start, end=ref_end+component_start)
                break

        if review_comment_ref:
            if quote_text is None:
                quote_text = component
            quote_start = quote_text.rfind('\n', 0, review_comment_ref.start) + 1
            quote_end = (quote_text + '\n').find('\n', review_comment_ref.end)
            line = lineno + quote_text[:quote_start + 1].count('\n')
            start = review_comment_ref.start - quote_start

            error = f"\x1B[1m{commit}:{line + 1}:{start + 1}: \x1B[31merror\x1B[39m: add context directly to commit messages instead of referring to review comments\x1B[m\n"
            last = 0
            for linem in re.finditer('\n', review_comment_ref.text + '\n'):
                if last == 0:
                    error += quote_text[quote_start:quote_start + start + linem.start() + 1] + '\n'
                    error += ' ' * start
                else:
                    error += review_comment_ref.text[last:linem.start()] + '\n'
                error += '\x1B[32m' + '^' * (linem.start() - last) + '\x1B[39m\n'

                last = linem.end()
            error += f"\x1B[1m{commit}:{line + 1}:{start + 1}: \x1B[30mnote\x1B[39m: prefer using --fixup when fixing previous commits in the same pull request\x1B[m"
            yield error


    if description is not None:
        yield from complain_about_excess_space(description)

        # Prevent upper casing the first letter of the first word, this is not a book-style sentence.
        title_case_re = regex.compile(r'\b[\p{Lu}\p{Lt}]\p{Ll}*(?:\s+[\p{Lu}\p{Lt}]\p{Ll}*)*\b')
        title_case_word = extract_match_group(title_case_re.match(description.text), 0, description.start)
        safe_words = (
                'TomTom',
            )
        if title_case_word and title_case_word.text not in safe_words:
            error = f"\x1B[1m{commit}:1:{title_case_word.start + 1}: \x1B[31merror\x1B[39m: don't use title case in the description\x1B[m\n"
            error += message.full_subject + '\n'
            error += ' ' * title_case_word.start + '\x1B[32m' + '^' + '~' * (title_case_word.end - title_case_word.start - 1) + '\x1B[39m'
            possibilities = difflib.get_close_matches(title_case_word.text, safe_words, n=1)
            if possibilities:
                error += f"\n\x1B[1m{commit}:1:{title_case_word.start + 1}: \x1B[30mnote\x1B[39m: did you mean {possibilities[0]!r}?\x1B[m"
            yield error

        yield from complain_about_review_refs(description.text, description.start, quote_text=message.full_subject)

        # Our own requirements on the description
        # No JIRA tickets in the subject line, because it wastes precious screen estate (80 chars)
        jira_tickets = []
        for m in JIRA_RE.finditer(description.text):
            jira_tickets.extend(range(*m.span()))
        if jira_tickets:
            error = "\x1B[1m{commit}:{line}:{col}: \x1B[31merror\x1B[39m: commit message's subject contains Jira tickets\x1B[m\n".format(line=1, col=description.start + 1 + jira_tickets[0], commit=commit)
            error += message.full_subject + '\n'
            error += '\x1B[32m'
            cur = -description.start
            for pos in jira_tickets:
                error += ' ' * (pos - cur) + '^'
                cur = pos + 1
            error += '\x1B[39m'
            yield error

        # Disallow ending the description with punctuation
        if re.match(r'.*[.!?,]$', description.text):
            error = f"\x1B[1m{commit}:1:{len(message.full_subject)}: \x1B[31merror\x1B[39m: commit message's subject ends with punctuation\x1B[m\n"
            error += message.full_subject + '\n'
            error += ' ' * (len(message.full_subject) - 1) + '\x1B[32m^\x1B[39m'
            yield error

        blacklist_start_words = (
                'added',
                'adds',
                'adding'
                'applied',
                'applies',
                'applying',
                'edited',
                'edits',
                'editing',
                'expanded',
                'expands',
                'expanding',
                'fixed',
                'fixes',
                'fixing',
                'removed',
                'removes',
                'removing',
                'renamed',
                'renames',
                'renaming',
                'deleted',
                'deletes',
                'deleting',
                'updated',
                'updates',
                'updating',
                'ensured',
                'ensures',
                'ensuring',
                'resolved',
                'resolves',
                'resolving',
                'verified',
                'verifies',
                'verifying',

                # repeating the tag is frowned upon as well
                type_tag.text,
            )
        blacklisted = extract_match_group(re.match(r'^(?:' + '|'.join(re.escape(w) for w in blacklist_start_words) + r')\b', description.text, flags=re.IGNORECASE), 0, description.start)
        if blacklisted:
            error = f"\x1B[1m{commit}:1:{blacklisted.start + 1}: \x1B[31merror\x1B[39m: commit message's description contains blacklisted word or repeats type tag\x1B[m\n"
            error += message.full_subject + '\n'
            error += blacklisted.start * ' ' + '\x1B[32m^' * (blacklisted.end - blacklisted.start) + '\x1B[39m\n'
            error += f"\x1B[1m{commit}:1:{blacklisted.start + 1}: \x1B[30mnote\x1B[39m: prefer using the imperative for verbs\x1B[m"
            yield error

    if len(message.subject) > 80:
        error = f"\x1B[1m{commit}:1:{81 + message.subject_start}: \x1B[31merror\x1B[39m: commit message's subject exceeds line length of 80 by {len(message.subject) - 80} characters\x1B[m\n"
        error += message.full_subject + '\n'
        error += ' ' * 79 + '\x1B[32m^' + '~' * (len(message.full_subject) - 80) + '\x1B[39m'
        yield error

    if len(message.lines) > 1 and message.lines[1]:
        error = f"\x1B[1m{commit}:2:1: \x1B[31merror\x1B[39m: commit message subject and body are not separated by an empty line\x1B[m\n"
        error += message.lines[1] + '\n'
        error += '\x1B[31m' + '~' * len(message.lines[1]) + '\x1B[39m'
        yield error

    # 8. Breaking changes MUST be indicated at the very beginning of the footer or body section of a commit. A breaking
    #    change MUST consist of the uppercase text BREAKING CHANGE, followed by a colon and a space.
    for paraidx, paragraph in enumerate(message.paragraphs):
        lineno = message.paragraph_line(paraidx)
        for m in re.finditer(r'\bBREAKING(\s+)CHANGE\b(\s*)(:?)(\s*)(\S?)', paragraph):
            if m.start() == 0:
                if paragraph[:m.end(4)] != 'BREAKING CHANGE: ' or not m.group(m.lastindex):
                    line_end = paragraph.find('\n')
                    error = f"\x1B[1m{commit}:{lineno + 1}:1: \x1B[31merror\x1B[39m: breaking changes should start with _exactly_ 'BREAKING CHANGE: ' and be followed by text immediately\x1B[m\n"
                    error += paragraph[:line_end] + '\n'
                    error += '\x1B[32m'
                    cur = 0
                    for group, expect in zip(range(1, m.lastindex + 1), (' ', '', ':', ' ')):
                        if m.group(group) != expect:
                            error += ' ' * (m.start(group) - cur)
                            error += '^' * len(m.group(group))
                            cur = m.end(group)
                    # 9. A description MUST be provided after the BREAKING CHANGE:, describing what has changed about the
                    #    API, e.g., BREAKING CHANGE: environment variables now take precedence over config files.
                    if not m.group(m.lastindex):
                        error += ' ' * (m.start(m.lastindex) - cur)
                        error += '^'
                        cur = m.start(m.lastindex) + 1
                    error += '\x1B[39m'
                    yield error
            else:
                line_start = paragraph.rfind('\n', 0, m.start()) + 1
                line_end = paragraph.find('\n', m.start(2))
                line = lineno + paragraph[:line_start + 1].count('\n')
                start = m.start() - line_start
                error = f"\x1B[1m{commit}:{line + 1}:{start + 1}: \x1B[31merror\x1B[39m: body contains 'BREAKING CHANGE' at other location than start of paragraph\x1B[m\n"
                error += paragraph[line_start:line_end] + '\n'
                error += '\x1B[32m' + ' ' * start
                error += ''.join('\n' if c == '\n' else '^' for c in paragraph[m.start():m.start(2)])
                error += '\x1B[39m'
                yield error

        yield from complain_about_review_refs(paragraph, lineno=lineno)


if __name__ == "__main__":
    from .cli import main
    sys.exit(main())



# Copyright (c) 2019 - 2020 TomTom N.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
import json
import os
import re
import sys
from textwrap import dedent

import pytest

try:
    # Python >= 3.8
    from importlib import metadata
except ImportError:
    import importlib_metadata as metadata


try:
    _hopic_version = tuple(
            (int(x) if re.match('^[0-9]+$', x) else x)
            for x in metadata.version('hopic').split('.')
        )
    hopic_cli = [ep for ep in metadata.entry_points()['console_scripts'] if ep.name == 'hopic'][0].load()
    from click.testing import CliRunner
    import git
except metadata.PackageNotFoundError:
    _hopic_version = ()

_git_time = f"{7 * 24 * 3600} +0000"


def run_with_config(config, args, files={}, env=None, cfg_file='hopic-ci-config.yaml'):
    runner = CliRunner(mix_stderr=False, env=env)
    with runner.isolated_filesystem():
        with git.Repo.init() as repo:
            if '/' in cfg_file and not os.path.exists(os.path.dirname(cfg_file)):
                os.makedirs(os.path.dirname(cfg_file))
            with open(cfg_file, 'w') as f:
                f.write(config)
            for fname, content in files.items():
                if '/' in fname and not os.path.exists(os.path.dirname(fname)):
                    os.makedirs(os.path.dirname(fname))
                with open(fname, 'w') as f:
                    f.write(content)
            repo.index.add((cfg_file,) + tuple(files.keys()))
            repo.index.commit(message='Initial commit', author_date=_git_time, commit_date=_git_time)
        if cfg_file != 'hopic-ci-config.yaml':
            args = ('--config', cfg_file) + tuple(args)
        result = runner.invoke(hopic_cli, args)

    if result.stdout_bytes:
        print(result.stdout, end='')
    if result.stderr_bytes:
        print(result.stderr, end='', file=sys.stderr)

    if result.exception is not None and not isinstance(result.exception, SystemExit):
        raise result.exception

    return result


@pytest.mark.skipif(_hopic_version < (1,15), reason="Hopic >= 1.15.0 not available")
def test_commisery_template(capfd):
    result = run_with_config(dedent('''\
                phases:
                  style:
                    commit-messages: !template "commisery"
                '''), ('show-config',))

    assert result.exit_code == 0
    output = json.loads(result.stdout, object_pairs_hook=OrderedDict)
    expanded = output['phases']['style']['commit-messages']
    assert expanded[0]['image'] is None
    commits, head = [e['sh'] for e in expanded]
    assert 'commisery.checking' in commits
    assert head[-2:] == ['commisery.checking', 'HEAD']


@pytest.mark.skipif(_hopic_version < (1,15), reason="Hopic >= 1.15.0 not available")
def test_commisery_template_ticket(capfd, monkeypatch):
    import hopic.build

    class MockGitInfo():
        target_commit = 'OUR_TARGET_COMMIT'

        @classmethod
        def from_repo(cls, *args):
            return cls()

    monkeypatch.setattr(hopic.build, 'HopicGitInfo', MockGitInfo)

    result = run_with_config(dedent('''\
                phases:
                  style:
                    commit-messages: !template
                      name: commisery
                      require-ticket: yes
                '''), ('show-config',))

    assert result.exit_code == 0
    output = json.loads(result.stdout, object_pairs_hook=OrderedDict)
    expanded = output['phases']['style']['commit-messages']
    assert expanded[0]['image'] is None
    commit_range, head = [e['sh'] for e in expanded]
    assert commit_range[-2:] == ['-j', 'OUR_TARGET_COMMIT..HEAD']
    assert 'commisery.checking' in commit_range
    assert head[-2:] == ['commisery.checking', 'HEAD']



# Copyright (c) 2019 - 2019 TomTom N.V. (https://tomtom.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..commit import *

import pytest


def test_basic_message_strip_and_splitup():
    """
    This tests these bits of functionality:

      * white space and comment stripping similar to how git-commit does it
      * line splitting
      * subject extraction
      * paragraph splitting
      * body extraction
    """
    m = CommitMessage('''\

# test stripping of comments and preceding empty lines

improvement(config): display config error messages without backtrace

In order to prevent users from thinking they're seeing a bug in Hopic.


This changes the type of ConfigurationError such that Click will display
its message without a backtrace. This ensures the displayed information
is more to the point.

# ------------------------ >8 ------------------------

This line and every other line after the 'cut' line above should not be present
in the output.

# test stripping of comments and succeeding empty lines

''')
    assert m.subject == '''improvement(config): display config error messages without backtrace'''
    assert m.lines[0] == m.subject

    assert m.paragraphs[0] == '''In order to prevent users from thinking they're seeing a bug in Hopic.'''
    assert m.paragraphs[0] == m.body.splitlines()[0]
    assert m.body.splitlines()[0] == m.message.splitlines()[2]

    assert m.paragraphs[1].splitlines(keepends=True)[0] == '''This changes the type of ConfigurationError such that Click will display\n'''

    assert m.paragraphs[-1].splitlines(keepends=True)[-1] == '''is more to the point.'''
    assert m.paragraphs[-1].splitlines()[-1] == m.body.splitlines()[-1]
    assert m.body.splitlines()[-1] == m.message.splitlines()[-1]

def test_commit_has_hexsha():
    commit = CommitMessage('fix: something', '0000000000000000000000000000000000000000')
    assert isinstance(commit.hexsha, str)

    conventional_commit = ConventionalCommit('fix: something', '0000000000000000000000000000000000000000')
    assert isinstance(conventional_commit.hexsha, str)

    reparsed_conventional_commit = CommitMessage(commit)
    assert isinstance(reparsed_conventional_commit.hexsha, str)

def test_conventional_missing_separator():
    with pytest.raises(RuntimeError, match=r'(lack|miss).*separator'):
        ConventionalCommit('fix display config error messages without backtrace')

    with pytest.raises(RuntimeError, match=r'(lack|miss).*separator'):
        ConventionalCommit('fix:display config error messages without backtrace')

def test_conventional_wrong_case_type_tag():
    with pytest.raises(RuntimeError, match=r'type.*tag.*upper.*case'):
        ConventionalCommit('Fix: display config error messages without backtrace')

def test_conventional_scoped_improvement():
    m = ConventionalCommit('improvement(config): display config error messages without backtrace')
    assert m.type_tag == 'improvement'
    assert m.scope == 'config'
    assert m.description == 'display config error messages without backtrace'
    assert not m.has_breaking_change()
    assert not m.has_new_feature()
    assert not m.has_fix()

def test_conventional_badly_scoped():
    with pytest.raises(RuntimeError, match=r'scope.*empty'):
        ConventionalCommit('fix(): display config error messages without backtrace')

    with pytest.raises(RuntimeError, match=r'scope.*whitespace'):
        ConventionalCommit('fix( config ): display config error messages without backtrace')

    with pytest.raises(RuntimeError, match=r'scope.*whitespace'):
        x = ConventionalCommit('fix(config error): display config error messages without backtrace')

def test_conventional_fix():
    m = ConventionalCommit('fix: use the common ancestor of the source and target commit for autosquash')
    assert m.type_tag == 'fix'
    assert m.scope is None
    assert not m.has_breaking_change()
    assert not m.has_new_feature()
    assert m.has_fix()

def test_conventional_new_feature():
    m = ConventionalCommit('''feat: make execution possible with 'hopic' as command''')
    assert m.type_tag == 'feat'
    assert m.scope is None
    assert not m.has_breaking_change()
    assert m.has_new_feature()
    assert not m.has_fix()

def test_conventional_break():
    m = ConventionalCommit('''\
chore: cleanup old cfg.yml default config file name

BREAKING-CHANGE: "${WORKSPACE}/cfg.yml" is no longer the default location
of the config file. Instead only "${WORKSPACE}/hopic-ci-config.yaml" is
looked at.
''')
    assert m.type_tag == 'chore'
    assert m.scope is None
    assert m.has_breaking_change()
    assert not m.has_new_feature()
    assert not m.has_fix()

    footer, = m.footers
    assert footer.token == 'BREAKING CHANGE'
    assert 'default location' in footer.value

def test_conventional_subject_break():
    m = ConventionalCommit('''chore!: delete deprecated 'ci-driver' command''')
    assert m.type_tag == 'chore'
    assert m.scope is None
    assert m.has_breaking_change()
    assert not m.has_new_feature()
    assert not m.has_fix()

def test_conventional_subject_breaking_fix():
    m = ConventionalCommit('''fix!: take parameter as unsigned instead of signed int''')
    assert m.type_tag == 'fix'
    assert m.scope is None
    assert m.has_breaking_change()
    assert not m.has_new_feature()
    assert m.has_fix()

def test_conventional_badly_marked_as_breaking():
    with pytest.raises(RuntimeError, match=r'breaking.*indicator'):
        ConventionalCommit('fix !: display config error messages without backtrace')

    with pytest.raises(RuntimeError, match=r'breaking.*indicator'):
        ConventionalCommit('fix! : display config error messages without backtrace')

    with pytest.raises(RuntimeError, match=r'breaking.*indicator'):
        ConventionalCommit('fix ! : display config error messages without backtrace')

def test_conventional_subject_breaking_new_feature():
    m = ConventionalCommit('''feat!: support multiple non-global current working directories''')
    assert m.type_tag == 'feat'
    assert m.scope is None
    assert m.has_breaking_change()
    assert m.has_new_feature()
    assert not m.has_fix()

def test_conventional_fixup_fix():
    m = ConventionalCommit('fixup! fix: only restore mtime for regular files and symlinks')
    assert m.type_tag == 'fix'
    assert m.scope is None
    assert m.description == 'only restore mtime for regular files and symlinks'
    assert not m.has_breaking_change()
    assert not m.has_new_feature()
    assert m.has_fix()


def test_multiple_fixups():
    """Permit multiple stacked fixup! and squash! prefixes"""
    m = CommitMessage('fixup! squash! fixup! something')
    assert m.autosquashed_subject == 'something'


def test_basic_footers():
    m = CommitMessage('''\
Merge #63: something

Bla bla

BREAKING CHANGE: something changed in an unpredicted way

Addresses #42 by working on finding the question
Implements: PIPE-123 through the obliviator
Acked-by: Alice <alice@example.com>
Merged-by: Hopic 1.21.2
Acked-by: Bob <bob@example.com>
''')
    with pytest.raises(KeyError):
        # git-trailer format for footers doesn't permit spaces in the token name
        _ = m.footers['BREAKING CHANGE']

    assert tuple(tuple(footer) for footer in m.footers) == (
            ('Addresses' , '#42 by working on finding the question'),
            ('Implements', 'PIPE-123 through the obliviator'),
            ('Acked-by'  , 'Alice <alice@example.com>'),
            ('Merged-by' , 'Hopic 1.21.2'),
            ('Acked-by'  , 'Bob <bob@example.com>'),
        )


def test_conventional_footers():
    m = ConventionalCommit('''\
Merge #63: improvement(groovy): retrieve execution graph in a single 'getinfo' call

This should reduce the amount of Jenkins master/slave interactions and
their associated Groovy script engine "context switches" (state
serialization and restoration). As a result performance should increase.

Addresses #279 by adding a test framework.

Acked-by: Anton Indrawan <Anton.Indrawan@tomtom.com>
Acked-by: Joost Muller <Joost.Muller@tomtom.com>
Acked-by: Martijn Leijssen <Martijn.Leijssen@tomtom.com>
Acked-by: Rene Kempen <Rene.Kempen@tomtom.com>
Merged-by: Hopic 0.10.2.dev7+g840ca0c
''')
    assert m.type_tag == 'improvement'
    assert m.scope == 'groovy'

    assert tuple(tuple(footer) for footer in m.footers) == (
            ('Addresses', '#279 by adding a test framework.'),

            ('Acked-by' , 'Anton Indrawan <Anton.Indrawan@tomtom.com>'),
            ('Acked-by' , 'Joost Muller <Joost.Muller@tomtom.com>'),
            ('Acked-by' , 'Martijn Leijssen <Martijn.Leijssen@tomtom.com>'),
            ('Acked-by' , 'Rene Kempen <Rene.Kempen@tomtom.com>'),

            ('Merged-by', 'Hopic 0.10.2.dev7+g840ca0c'),
        )



#!/usr/bin/env python3

# Copyright (c) 2020 - 2020 TomTom N.V. (https://tomtom.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import logging
import os
import re
import subprocess
import sys

import click
import click_log

from commisery.checking import (
    check_commit,
    DEFAULT_ACCEPTED_TAGS,
)
from commisery.range import check_commit_rev_range


log = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)
log.addHandler(logging.NullHandler())


@click.command()
@click.argument('target', nargs=-1)
@click.option('--ticket', '-j',
              default=None,
              is_flag=True,
              multiple=True,
              help='When provided, the commit messages in the provided revision range '
                   + 'must contain at least one ticket reference, in the form of regex:\n'
                   + '\\b[A-Z]+-[0-9]+\\b'
              )
@click.option('--tags', '-t',
              multiple=True,
              help='Comma-separated list of accepted conventional commit tags to allow '
                   + 'aside from the default "feat" and "fix".\n'
                   + f'If omitted, uses the following list:\n{", ".join(DEFAULT_ACCEPTED_TAGS)}'
              )
@click_log.simple_verbosity_option(__package__.split('.')[0])
def main(target, tags, ticket):
    """
    Checks whether commit messages adhere to the Convention Commits standard.

    \b
    TARGET can be:
     - a single commit hash
     - a file containing the commit message to check
     - a revision range that `git rev-list` can interpret
    When TARGET is omitted, 'HEAD' is implied.

    Optional parameters may be provided to specify which tags are allowed
    and/or whether the commits in the revision range should contain a ticket reference.
    """

    custom_accepted_tags = None if not tags else (frozenset(itertools.chain.from_iterable(x.split(',') for x in tags)))
    target = target or ('HEAD',)
    target_str = ' '.join(target)

    # We will treat the input as a revision range if:
    # the target is not a full SHA-1 hash (i.e. 40 hex digits) and does not represent an existing file
    # and the result of a `git rev-parse <target>` yields more than one line
    try:
        if not re.match(r'^[0-9a-fA-F]{40}$', target_str) and \
           not os.path.exists(target_str) and \
           len(subprocess.check_output(('git', 'rev-parse') + target,
                                       stderr=subprocess.DEVNULL).decode(encoding='UTF-8').splitlines()) > 1:
            log.debug(f'Handling as range: {target_str}')
            result = check_commit_rev_range(target, custom_accepted_tags=custom_accepted_tags, require_ticket=ticket)
        else:
            log.debug(f'Handling as commitish: {target_str}')
            result = check_commit(target_str, custom_accepted_tags=custom_accepted_tags, require_ticket=ticket)
    except subprocess.CalledProcessError:
        result = 1
        log.error('Could not parse target revision spec %s', target_str)

    sys.exit(result)


if __name__ == '__main__':
    sys.exit(main())



# Copyright (c) 2018 - 2020 TomTom N.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup
import re

with open('README.md', encoding='UTF-8') as fh:
    long_description = fh.read()

# Extract first paragraph
description = re.sub(r'^\s*(?:#+[^\n]*\s*)*', r'', long_description, count=1, flags=re.DOTALL|re.MULTILINE)
description = re.sub(r'\n\n.*', r'', description, flags=re.DOTALL|re.MULTILINE)
# Eliminate link annotation
description = re.sub(r'\[([^\]]+)\]', r'\1', description)
# Convert line breaks into spaces
description = description.replace('\n', ' ')

setup(
    name='commisery',
    author='TomTom N.V.',
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=('commisery','commisery.cli'),
    py_modules=('commisery',),
    python_requires='>=3.6.5',
    install_requires=(
      'Click>=7.1.2,<8',
      'click-log',
      'GitPython>=3,<4',
      'regex',
      'stemming>=1,<2',
    ),
    setup_requires=(
      'pytest-runner',
      'setuptools_scm',
      'setuptools_scm_git_archive',
    ),
    tests_require=(
      'pytest',
    ),
    extras_require={
      'github': [
        'PyGithub>=1.53,<2',
      ],
    },
    use_scm_version={"relative_to": __file__},
    entry_points={
      'console_scripts': [
        'commisery-verify-msg = commisery.cli:main',
        'commisery-verify-github-pullrequest = commisery.github:main [github]',
      ],
      'hopic.plugins.yaml': [
        'commisery = commisery.hopic_template:commisery',
      ],
    },
    zip_safe=True,
    url='https://github.com/tomtom-international/commisery',
    project_urls={
      'Source Code': 'https://github.com/tomtom-international/commisery',
    },
    classifiers=(
      'License :: OSI Approved :: Apache Software License',
    ),
    license='Apache License 2.0',
    license_file='LICENSE',
)



#!/usr/bin/env python3

# Copyright (c) 2020 - 2020 TomTom N.V. (https://tomtom.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import git

from commisery.checking import (
    check_commit,
    count_ticket_refs,
)


log = logging.getLogger(__name__)


def check_commit_rev_range(revision_range, custom_accepted_tags=None, require_ticket=False):
    """
    Checks commit messages of the commits in the provided revision range.

    Parameters:
        revision_range: a sequence of `git rev-list`-compatible arguments describing a revision range
        custom_accepted_tags: an optional list of conventional commit tags to allow besides 'feat' and 'fix'
        require_ticket: when set to true, requires a Jira-style ticket to be present anywhere in the commit message
                in one of the commits described by revision_range

    Returns:
        1 on failure, 0 on success.
    """
    log.debug('Revision range: %s', ' '.join(revision_range))
    if custom_accepted_tags:
        log.debug('Custom accepted tags: %s', ' '.join(custom_accepted_tags))

    try:
        with git.Repo(search_parent_directories=True) as repo:
            try:
                commits = list(repo.iter_commits(rev=revision_range))
            except git.GitCommandError:
                log.exception('Error getting list of commits from revision range: %s', ' '.join(revision_range))
                return 1

            log.debug('Repo at %s has %d commits for range: %s',
                      repo.working_dir, len(commits),  ' '.join(revision_range))
            ticket_count = 0
            error_count = 0

            for commit in commits:
                log.debug('Checking commit: %s', commit.hexsha)
                error_count += check_commit(commit.hexsha, custom_accepted_tags=custom_accepted_tags)
                if require_ticket:
                    ticket_count += count_ticket_refs(commit.message)

            log.debug('Done checking commits{}{}'.format(
                f', {error_count} error{"s" if error_count > 1 else ""} found' if error_count else '',
                f', ticket reference count: {ticket_count}' if require_ticket else ''))

            if require_ticket and ticket_count == 0:
                log.error('No ticket is referenced within the provided range of commits')
                error_count += 1

            if error_count == 0:
                return 0

    except (git.InvalidGitRepositoryError, git.NoSuchPathError):
        log.error('Not a git repository')

    return 1
